/* Bounds Checking for GCC.
   Copyright (C) 1995 Richard W.M. Jones <rjones@orchestream.com>.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   File:
  	lib/alloca.c
   Summary:
  	Implementation of `alloca' function. If you compile (as recommended)
  	with -fno-builtin, you will get this implementation of a checked
  	alloca, with optional checks for stale pointers and so forth.
   Other notes:
  	
   Author      	Date		Notes
   RWMJ		31/3/95		Initial implementation.
*/ 

#include <stdio.h>
#include <stdlib.h>

#include "bounds-lib.h"

#include "mutex.h"			/* for 'enter|leave_critical_section'*/

#if defined(__BOUNDS_CHECKING_ON)
#error "This file must not be compiled with bounds checking enabled."
#endif

#if defined(alloca)
#undef alloca
#endif

#if defined(__alloca)
#undef __alloca
#endif

static __inline__ void *
bounds_alloca (const char *filename, int line, size_t size)
{
  alloca_item *item;
  sigset_t old_mask;

  enter_critical_section (&old_mask);
  item = __bounds_malloc (sizeof (alloca_item));
  if (!item)
    __bounds_internal_error ("out of space allocating alloca chain",
			     __FILE__, __LINE__);

  /* Get `functions.c' to add the current alloca item to the current function
     context. When the current function exits, we will be called (in
     __bounds_alloca_free) with a chain of these allocations.  */
  __bounds_add_alloca_chain (item);

  /* Try to allocate the space as a checked memory allocation. Note:
     (1) alloca never fails,
     (2) `free' will be called for this object, but according to the
     current heap mode, it may not actually be freed, and
     (3) we allocate from the heap, not the stack.  */
  /* We do not support zero sized objects.  */
  if (size == 0)
    size = 1;
  item->pointer = __bounds_malloc (size + 1);
  item->filename = filename;
  item->line = line;
  if (! item->pointer)
    __bounds_internal_error ("alloca failed (out of memory)", filename, line);
  __bounds_add_heap_object (item->pointer, size, 1, "alloca", 0,
			    filename, line);
  leave_critical_section (&old_mask);

  return item->pointer;
}

void *
__alloca (size_t size)
{
  return bounds_alloca (NULL, 0, size);
}

void *
alloca (size_t size)
{
  return bounds_alloca (NULL, 0, size);
}

void *
__bounds_check_alloca (const char *filename, int line, size_t size)
{
  return bounds_alloca (filename, line, size);
}

void *
__bounds_check___alloca (const char *filename, int line, size_t size)
{
  return bounds_alloca (filename, line, size);
}

void *
__bounds_check___builtin_alloca (const char *filename, int line, size_t size)
{
  return bounds_alloca (filename, line, size);
}

/* Free up the alloca stack for the current function.  */

void
__bounds_alloca_free (alloca_item *alloca_stack)
{
  alloca_item *next;

  while (alloca_stack)
    {
      next = alloca_stack->next;
      __bounds_delete_heap_object(alloca_stack->pointer, -1,
				  "invalid pointer used in `alloca'",
				  alloca_stack->filename, alloca_stack->line);
      __bounds_free (alloca_stack->pointer);
      __bounds_free (alloca_stack);
      alloca_stack = next;
    }
}
