/* Bounds Checking for GCC.
   Copyright (C) 1995 Richard W.M. Jones <rjones@orchestream.com>.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   File:
  	lib/functions.c
   Summary:
  	Deal with pushing and popping of function contexts and parameters
  	of those function contexts.
   Other notes:
  	
   Author      	Date		Notes
   RWMJ		6/3/95		Initial implementation.
   RWMJ		3/4/95		Some optimizations.
*/

#include <stdio.h>
#include <stdlib.h>

#include "bounds-lib.h"

#include "mutex.h"			/* for 'enter|leave_critical_section'*/
#include "check.h"

#if defined(__BOUNDS_CHECKING_ON)
#error "This file must not be compiled with bounds checking enabled."
#endif

cur_context *__bounds_function_context = NULL;
static cur_context *__bounds_function_list = NULL;

/* Helper functions that allocate and deallocate function contexts.
   We used to allocate and free them immediately, but that was
   quite inefficient, so now we allocate blocks of function contexts
   and never free them. (RWMJ, 3/4/95).  */

static __inline__ cur_context *
get_context_no_lock (void)
{
  int current_thread;
  cur_context *current_context;

  if (__bounds_function_thread)
    current_thread = (*__bounds_function_thread) ();
  else
    current_thread = 0;
  current_context = __bounds_function_context;
  if (current_context
      && current_context->current_thread != current_thread)
    {
      current_context = __bounds_function_list;
      while (current_context)
	{
	  if (current_context->current_thread == current_thread)
	    break;
	  current_context = current_context->next;
	}
      __bounds_function_context = current_context;
    }
  if (current_context == NULL)
    {
      current_context = __bounds_malloc (sizeof (cur_context));
      if (!current_context)
	__bounds_internal_error ("out of memory allocating a block"
				 " of function contexts",
				 __FILE__, __LINE__);
      current_context->current_thread = current_thread;
      current_context->nesting_nr = 0;
      current_context->func_context.next =
      current_context->func_context.prev = &current_context->func_context;
      current_context->hist_func = NULL;
      current_context->last_end = NULL;
      current_context->func_context.no_checking = 0;
      current_context->next = __bounds_function_list;
      __bounds_function_list = current_context;
      __bounds_function_context = current_context;
    }
  return current_context;
}

static __inline__ cur_context *
get_context (void)
{
  cur_context *current_context;
  sigset_t old_mask;

  enter_critical_section (&old_mask);
  current_context = get_context_no_lock ();
  leave_critical_section (&old_mask);
  return current_context;
}

/* Return a pointer to a thread safe strtok pointer.  */

char **
__bounds_strtok_last_end (void)
{
  cur_context *cc = get_context ();
  return &cc->last_end;
}

static __inline__ function_context *
allocate_new_context (cur_context *cc)
{
  function_context *fc;
  sigset_t old_mask;

  if (cc->hist_func)
    {
      fc = cc->hist_func;
      cc->hist_func = fc->next;
    }
  else
    {
      enter_critical_section (&old_mask);
      fc = __bounds_malloc (sizeof (function_context));
      leave_critical_section (&old_mask);
      if (fc == NULL)
	__bounds_internal_error ("out of memory allocating a function contexts",
				 __FILE__, __LINE__);
      /* Mark a function context so we can find it when deleting a spurious
	 stack object.  */
      fc->parms.base = fc->stack.base = ILLEGAL;
    }

  fc->next = cc->func_context.next;
  fc->prev = &cc->func_context;
  fc->next->prev = fc;
  cc->func_context.next = fc;

  return fc;
}

static __inline__ void
remove_this_context (cur_context *cc, function_context *fc)
{
  /* Never realy free function context. Put it in a history list.  */
  fc->prev->next = fc->next;
  fc->next->prev = fc->prev;
  fc->next = cc->hist_func;
  cc->hist_func = fc;
}

/* Pop out of a function context.  */

static __inline__ void
__bounds_internal_pop_function (const char *name, int nesting_nr)
{
  int current_nesting_nr;
  const char *context_name;
  cur_context *cc;
  function_context *fc;
  object *obj, *nobj;

  cc = get_context_no_lock ();
  for (;;)
    {
      fc = cc->func_context.next;
      if (fc == &cc->func_context)
        __bounds_internal_error ("too many function context pops",
			         __FILE__, __LINE__);
      current_nesting_nr = fc->nesting_nr;
      context_name = fc->name;
      /* Delete all the parameters to this function.  */
      obj = fc->parms.next;
      while (obj != &fc->parms)
	{
	  nobj = obj->next;
          __bounds_internal_delete_stack_object (obj, -1);
	  obj = nobj;
	}
      /* Delete all the stack parameters in this function.  */
      obj = fc->stack.next;
      while (obj != &fc->stack)
	{
	  nobj = obj->next;
          __bounds_internal_delete_stack_object (obj, -1);
	  obj = nobj;
	}

      /* Call the alloca module to (if necessary) delete items from the
         alloca stack. Only do this if `alloca_stack' is non-NULL to try
         to avoid the function call.  */
      if (fc->alloca_stack != NULL)
        __bounds_alloca_free (fc->alloca_stack);

      /* Delete the function_context structure.  */
      remove_this_context (cc, fc);
      if (nesting_nr == current_nesting_nr)
        break;
    }
  cc->nesting_nr = current_nesting_nr;

  /* Check that the name passed matches the name pushed into. If not, this
     is probably a compiler error, else longjmp was used.  */
  if (context_name != name &&
      __bounds_strcmp (context_name,name) != 0)
    __bounds_internal_error ("popped out of a mismatched function context",
			     __FILE__, __LINE__);

  /* If we aren't checking bounds in this function context, switch off
     bounds checking now.  */
  __bounds_debug_no_checking = cc->func_context.next->no_checking;
}

/* Add a stack object to the function list nesting_nr.
   This call is made when we are inside a critical section.  */

void
__bounds_add_function_object (object *obj, int nesting_nr)
{
  cur_context *cc = get_context_no_lock ();
  function_context *fc = cc->func_context.next;

  if (fc == &cc->func_context)
    __bounds_internal_error ("__bounds_add_function_object called when"
			     " not in a function context",
			     __FILE__, __LINE__);

  /* If not at correct level pop context until we are.  */
  while (fc->nesting_nr != nesting_nr)
    {
      __bounds_internal_pop_function (fc->name, fc->nesting_nr);
      fc = cc->func_context.next;
    }

  if (fc == &cc->func_context)
    __bounds_internal_error ("__bounds_add_function_object called when"
			     " not in a function context",
			     __FILE__, __LINE__);

  /* Make a reverse list. This will make remove_function_object faster.  */
  obj->next = fc->stack.next;
  fc->stack.next = obj;
}

/* Delete a stack object of the function list nesting_nr.
   This call is made when we are inside a critical section.  */

int
__bounds_remove_function_object (object *obj, int nesting_nr)
{
  cur_context *cc = get_context_no_lock ();
  function_context *fc = cc->func_context.next;
  object *nobj, *lobj;

  if (fc == &cc->func_context)
    __bounds_internal_error ("__bounds_remove_function_object called when"
			     " not in a function context",
			     __FILE__, __LINE__);

  /* If not at correct level pop context until we are.  */
  while (fc->nesting_nr != nesting_nr)
    {
      __bounds_internal_pop_function (fc->name, fc->nesting_nr);
      fc = cc->func_context.next;
    }

  if (fc == &cc->func_context)
    __bounds_internal_error ("__bounds_remove_function_object called when"
			     " not in a function context",
			     __FILE__, __LINE__);

  /* Walk through the list. It is most likely that the first
     object is our target cause we add them in reverse order
     in the list.  */
  nobj = fc->stack.next;
  if (obj == nobj)
    {
      fc->stack.next = nobj->next;
    }
  else
    {
      lobj = nobj;
      nobj = nobj->next;
      while (nobj != &fc->stack)
	{
	  if (obj == nobj)
	    {
	      lobj->next = nobj->next;
	      return 1;
	    }
	  lobj = nobj;
	  nobj = nobj->next;
        }
      nobj = fc->parms.next;
      if (obj == nobj)
        {
          fc->parms.next = nobj->next;
        }
      else
        {
          lobj = nobj;
          nobj = nobj->next;
          while (nobj != &fc->parms)
	    {
	      if (obj == nobj)
	        {
	          lobj->next = nobj->next;
	          return 1;
	        }
	      lobj = nobj;
	      nobj = nobj->next;
            }
          return 0;
	}
    }
  return 1;
}

/* Delete a function context when a spurious stack object is found. See
   also function __bounds_internal_add_stack_object in objects.c
   This call is made when we are inside a critical section.  */

void
__bounds_remove_function_context (object *obj, int nesting_nr)
{
  cur_context *cc;
  function_context *fc;
  object *nobj = obj;

  /* Find current function context.  */
  while (nobj->base != ILLEGAL)
    nobj = nobj->next;
  cc = (cur_context *)nobj->parent;
  fc = (function_context *)nobj->left;

  if (fc->nesting_nr == nesting_nr
      && __bounds_remove_function_object (obj, nesting_nr))
    {
      __bounds_internal_delete_stack_object (obj, -1);
      return;
    }
#if DEBUG_FEATURES
  if (__bounds_debug_print_calls)
    printf ("__bounds_remove_function_context (name=\"%s\", nesting_nr=%d)\n",
	    fc->name, fc->nesting_nr);
#endif

  /* Delete all the parameters to this function.  */
  obj = fc->parms.next;
  while (obj != &fc->parms)
    {
      nobj = obj->next;
      __bounds_internal_delete_stack_object (obj, -1);
      obj = nobj;
    }

  /* Delete all the stack parameters in this function.  */
  obj = fc->stack.next;
  while (obj != &fc->stack)
    {
      nobj = obj->next;
      __bounds_internal_delete_stack_object (obj, -1);
      obj = nobj;
    }

  /* Delete alloca objects.  */
  if (fc->alloca_stack != NULL)
    __bounds_alloca_free (fc->alloca_stack);

  if (cc->func_context.next == fc)
    cc->nesting_nr = fc->nesting_nr;

  remove_this_context (cc, fc);
}

/* Push into a function context.  */

int
__bounds_push_function (const char *name, int in_main,
			const char *filename, int line)
{
  cur_context *cc = get_context ();
  function_context *fc;

#if COLLECT_STATS
  ++__bounds_stats_push_function;
#endif
#if DEBUG_FEATURES
  if (__bounds_debug_print_calls)
    printf ("__bounds_push_function"
	    "(name=\"%s\", main=%d, file=\"%s\", ln=%d), nesting_nr=%d\n",
	    name, in_main, filename, line, cc->nesting_nr);
#endif

  if (in_main)
    {
      static int count_main = 0;

      /* This function is supposedly `main'. `main' should occur once and only
         once. ANSI doesn't allow `main' to be called recursively. Notice that
         `main' is not usually the first function to be called. That is usually
         one of our global constructor functions.  */
      count_main ++;

      if (count_main > 1)
        {
  	  in_main = 0;	/* Not really in main, if called recursively.  */
  	  if (count_main == 2)
	    __bounds_warning (filename, line, NULL,
			      "ANSI does not allow main ()"
			      " to be called recursively");
        }
    }

  /* Allocate space for the function_context structure plus all the
     arguments we're going to get.  */
  fc = allocate_new_context (cc);

  /* Set the context variables as appropriate.  */
  fc->nesting_nr = cc->nesting_nr;
  cc->nesting_nr++;
  fc->in_main = in_main;
  fc->no_checking = 0;
  fc->name = name;
  fc->filename = filename;
  fc->line = line;
  fc->parms.next = &fc->parms;
  fc->stack.next = &fc->stack;
  fc->alloca_stack = NULL;
  /* Make base special so we can find the current context and 
     function context pointers when we have to delete a stack object.  */
  fc->parms.parent = fc->stack.parent = (object *)cc;
  fc->parms.left = fc->stack.left = (object *)fc;

  /* Switch on checking for this function by default. It may be switched off
     by a call to `__bounds_no_checking_in_this_function'.  */
  __bounds_debug_no_checking = 0;
  return fc->nesting_nr;
}

/* Pop out of a function context.  */

void
__bounds_pop_function (const char *name, int nesting_nr)
{
  sigset_t old_mask;

#if COLLECT_STATS
  ++__bounds_stats_pop_function;
#endif


#if DEBUG_FEATURES
  if (__bounds_debug_print_calls)
    printf ("__bounds_pop_function (name=\"%s\", nesting_nr=%d)\n",
	    name, nesting_nr);
#endif

  enter_critical_section (&old_mask);
  __bounds_internal_pop_function (name, nesting_nr);
  leave_critical_section (&old_mask);
}

/* Add a parameter. This function may be called repeatedly as we
   enter a function context.  */

void
__bounds_add_param_object (void *ptr, size_t size, size_t align,
			   const char *filename, int line, const char *name,
			   int nesting_nr)
{
  sigset_t old_mask;
  cur_context *cc;
  function_context *fc;
  object *obj;

#if COLLECT_STATS
  ++__bounds_stats_param_function;
#endif

#if DEBUG_FEATURES
  if (__bounds_debug_print_calls)
    printf ("__bounds_add_param_object"
	    "(p=%p, sz=%u, align=%u, file=\"%s\", ln=%d, name=\"%s\""
	    " nesting_nr=%d)\n",
	    ptr, size, align, filename, line, name, nesting_nr);
#endif

  enter_critical_section (&old_mask);
  cc = get_context_no_lock ();
  fc = cc->func_context.next;
  if (fc == &cc->func_context)
    __bounds_internal_error ("__bounds_add_param_object called when"
			     " not in a function context",
			     __FILE__, __LINE__);

  /* Add the stack object, and note it down in the function_context, so
     we can automatically delete it later. We set the `no padding' flag
     in the object since padding isn't added between parameters for
     compatibility reasons.  */
  obj = __bounds_internal_add_stack_object (ptr, size, align,
					    filename, line, name,
					    1, -1);
  obj->next = fc->parms.next;
  fc->parms.next = obj;

  /* If this is `main' and we have the first two arguments, then note the
     positions of the command line arguments here.  */
  if (fc->in_main && obj && obj->next && obj->next->next == &fc->parms)
    {
      void *arg0_ptr, *arg1_ptr;
      int argc;
      char **argv;

      /* points to `argc'.  */
      arg0_ptr = obj->next->base;
      /* points to `argv'.  */
      arg1_ptr = obj->base;

      argc = *(int *) arg0_ptr;			/* get them */
      argv = *(char ***) arg1_ptr;

      __bounds_note_main_args (argc, argv);
    }
  leave_critical_section (&old_mask);
}

/* If you don't want to check bounds in the currently active function,
   then call this somewhere near the beginning of the function.  */

void
__bounds_no_checking_in_this_function (void)
{
  cur_context *cc = get_context ();
  function_context *fc = cc->func_context.next;
  if (fc != &cc->func_context)
    {
      fc->no_checking = 1;
      __bounds_debug_no_checking = 1;
    }
}

/* Print a stack trace. This function is of little interest, since
  (a) it misses out all unchecked functions, and
  (b) GDB does a better stack trace already.  */
 
void
__bounds_debug_stack_trace (void)
{
  cur_context *cc = get_context_no_lock ();
  function_context *fc = cc->func_context.next;
  object *obj;

  if (fc == &cc->func_context)
    {
      printf ("No stack.\n");
      return;
    }

  printf ("Stack trace (params are reversed):\n");
  while (fc != &cc->func_context)
    {
      printf ("- %s (", fc->name);
      obj = fc->parms.next;
      while (obj != &fc->parms)
	{
	  if (obj != fc->parms.next) printf (", ");
	  printf ("%s = ", obj->name);
	  /* We can only print the argument value if it's an int or a
	     pointer.  */
	  if (obj->size == sizeof (int)
	      && obj->align == sizeof (int))
	    printf ("%u", *(unsigned *) (obj->base));
	  else
	    printf ("?");
	  obj = obj->next;
	}
      printf (") at %s:%d\n", fc->filename, fc->line);
      fc = fc->next;
    }
}

/* The following function is called from the `alloca.c' code to
   update the alloca chain in the current function context. Placing
   this function here adds some modularity at the expense of some
   speed. Returns the current function name. */

void
__bounds_add_alloca_chain (alloca_item *item)
{
  cur_context *cc = get_context_no_lock ();
  function_context *fc = cc->func_context.next;

  if (fc != &cc->func_context)
    {
      item->next = fc->alloca_stack;
      fc->alloca_stack = item;
    }
  else
    {
      /* Allow this because we can have mixed objects from bounds checked code
         and non bounds checked code. This also means that the item will never
         be freed.  */
#if 0
      __bounds_internal_error ("added item to alloca stack in NULL"
			       " function context",
			       __FILE__, __LINE__);
#else
#endif
    }
}
