/* Bounds Checking for GCC.
   Copyright (C) 1995 Richard W.M. Jones <rjones@orchestream.com>.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   File:
  	lib/print.c
   Summary:
  	Simple printf/vprintf implementations that are guaranteed not to
  	use malloc. (Calling malloc while printing an error message might
  	cause recursion).
   Other notes:
  	The following specifiers are implemented:
  		%p	pointer, prints NULL, ILLEGAL or value (as hex)
  		%d	signed integer
  		%u	unsigned integer
  		%s	string, prints "(null)" if NULL pointer passed
  		%x	hexadecimal integer
   Author      	Date		Notes
   RWMJ		5/12/94		Initial implementation.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>

#include "bounds-lib.h"

#if defined(__BOUNDS_CHECKING_ON)
#error "This file must not be compiled with bounds checking enabled."
#endif

/* Note: 5/10/95. OSF/1 3.0 appears to allow reentrant calls to printf and
   friends. So this isn't needed ...  */

#if !defined(__osf__)

#if defined(printf)
#undef printf
#endif
#if defined(vprintf)
#undef vprintf
#endif

/* Low level functions to manage a simple output buffer.  */

static char buffer [256], *next = buffer;

static __inline__ void
printchar (char c)
{
  *next++ = c;
  if (c == '\n' || next - buffer == 256)
    {
      /* Flush output buffer.  */
      write (2, buffer, next - buffer);
      next = buffer;
    }
}

/* low level functions to convert numbers to strings.  */

static __inline__ char *
_ul_to_a (unsigned long u, char *buf)
{
  if (u != 0)
    {
      unsigned long n = u % 10;
      buf = _ul_to_a (u / 10, buf);
      *buf++ = n + '0';
    }
  return buf;
}

static __inline__ void
l_to_a (long i, char *buf)
{
  if (i == 0)
    *buf++ = '0';
  else
    {
      if (i < 0)
	{
    	  *buf++ = '-';
    	  i = -i;
    	}
      buf = _ul_to_a ((unsigned long) i, buf);
    }
  *buf = 0;
}

static __inline__ void
ul_to_a (unsigned long u, char *buf)
{
  if (u == 0)
    *buf++ = '0';
  else
    buf = _ul_to_a (u, buf);
  *buf = 0;
}

/* Print a hex number. `width' is the number of digits (ie. number of
   hex nybbles) to print.  */

static __inline__ void
xtoa (unsigned long u, char *buf, int width)
{
  static char hexdigit[16] = {'0', '1', '2', '3', '4', '5',
			      '6', '7', '8', '9', 'a', 'b',
			      'c', 'd', 'e', 'f'};
  int i;
  unsigned d;
  unsigned long top_nybble_mask = 0xF << ((width - 1) * 4);

  for (i = width-1; i >= 1; --i)
    {
      if (u & top_nybble_mask)
        break;
      u <<= 4;
    }
  for (; i >= 0; --i)
    {
      d = u >> ((width - 1) * 4);
      u <<= 4;
      *buf++ = hexdigit[d];
    }
  *buf = 0;
  return;
}

/* Like vprintf.  */

void
__bounds_vprintf (const char *format, va_list args)
{
  const char *p;
  char *s;
  int i;
  int prec;
  char pchar;
  unsigned u;
  void *vp;
  char buf [64];

  __bounds_debug_no_checking = 1;
  p = format;
  while (*p)
    {
      if (*p != '%')
	{
          printchar (*p);
          p++;
        }
      else
	{
          p++;
          prec = 0;
          pchar = ' ';
          if (*p == '0') pchar = *p++;	/* check for '%02d'    */
          while (*p >= '0' && *p <= '9')/* calculate precision */
            prec = prec * 10 + *p++ - '0';
          switch (*p)
	    {
            case 's':
	      /* String. Print it, or print "(null)" if the pointer passed is
	         NULL.  */
	      s = va_arg (args, char *);
	      if (s == NULL)
		{
          	  prec -= __bounds_strlen ("(null)");
          	  while (prec-- > 0) printchar (pchar);
	  	  __bounds_printf ("(null)");
		}
	      else
		{
          	  prec -= __bounds_strlen (s);
          	  while (prec-- > 0) printchar (pchar);
	  	  while (*s)
		    {
	    	      printchar (*s);
	    	      ++s;
	  	    }
		}
	      break;
            case 'd':
	      /* Signed integer.  */
	      i = va_arg (args, int);
	      l_to_a (i, buf);
              prec -= __bounds_strlen (buf);
              while (prec-- > 0) printchar (pchar);
	      __bounds_printf ("%s", buf);
	      break;
            case 'u':
	      u = va_arg (args, unsigned);
	      ul_to_a (u, buf);
              prec -= __bounds_strlen (buf);
              while (prec-- > 0) printchar (pchar);
	      __bounds_printf ("%s", buf);
	      break;
            case 'p':
	      /* Pointer. Print NULL, ILLEGAL or a representation of the
		 pointer.  */
	      vp = va_arg (args, void *);
	      if (vp == NULL)
		{
          	  prec -= __bounds_strlen ("NULL");
          	  while (prec-- > 0) printchar (pchar);
	  	  __bounds_printf ("NULL");
		}
	      else if (vp == ILLEGAL)
		{
          	  prec -= __bounds_strlen ("ILLEGAL");
          	  while (prec-- > 0) printchar (pchar);
	  	  __bounds_printf ("ILLEGAL");
		}
	      else
		{
	  	  xtoa (PTR_TO_UNSIGNED (vp), buf,
			sizeof (ptr_as_unsigned_t) * 2);
          	  prec -= __bounds_strlen ("0x") + __bounds_strlen (buf);
          	  while (prec-- > 0) printchar (pchar);
	  	  __bounds_printf ("0x%s", buf);
		}
	      break;
            case 'x':
	      /* Hexadecimal integer.  */
	      u = va_arg (args, unsigned);
	      xtoa (u, buf, sizeof (unsigned) * 2);
              prec -= __bounds_strlen (buf);
              while (prec-- > 0) printchar (pchar);
	      __bounds_printf ("%s", buf);
	      break;
            case '%':
	      printchar ('%'); break;
            case 0:
	      printchar ('\\'); printchar ('0'); break;
            default:
	      /* Unrecognized format specifier. Print character and advance va
	         pointer along one to ignore that argument.  */
	      printchar (*p);
	      (void) (va_arg (args, int));
            }
          if (*p) p++;
        }
    }
  __bounds_debug_no_checking = 0;
}

/* Like printf.  */

void
__bounds_printf (const char *format, ...)
{
  va_list args;

  va_start (args, format);
  __bounds_vprintf (format, args);
  va_end (args);
}

#endif /* !defined (__osf__) */
