/*----------------------------------------------------------------------*
 * Bounds Checking for GCC.                                             *
 * Copyright (C) 1995 Richard W.M. Jones <rjones@orchestream.com>.      *
 *----------------------------------------------------------------------*
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
 *----------------------------------------------------------------------*
 * File:
 *      thread/pthread_init.c
 * Summary:
 *      Init bounds checking for pthread library or signal handlers.
 * Other notes:
 *	Signal handlers are not turned on by default because it takes
 *	a lot of cpu cycles. Most of the time this extra overhead is
 *	not needed because the user code does not have signal handlers
 *	or the user signal handlers do not call bounds_checking functions.
 * Author       Date            Notes
 * HtB         2/6/99         Initial implementation.
 *----------------------------------------------------------------------*/

#ifdef __BOUNDS_CHECKING_ON

/* If signal handlers are used set this define to '1' instead of '0'.  */

#define	USE_SIGNAL	0

/* If pthread is used set this define to '1' instead of '0'.  */

#define	USE_PTHREAD	1

#include <stdio.h>		/* NULL */
#include <sys/types.h>
#if USE_PTHREAD
#include <pthread.h>
#endif

#if USE_SIGNAL
extern int __bounds_mutex_signal;
#endif

#if USE_PTHREAD
extern void *(*__malloc_hook)(size_t);
extern void *(*__realloc_hook)(void *, size_t);
extern void (*__free_hook)(void *);

extern void *__bounds_malloc(size_t);
extern void *__bounds_realloc(void *,size_t);
extern void __bounds_free(void *);

static void *bounds_malloc_hook(size_t size);
static void *bounds_realloc_hook(void *ptr, size_t size);
static void bounds_free_hook(void *ptr);

static void bounds_mutex_start(void);
static void bounds_mutex_end(void);

extern pthread_t (*__bounds_function_thread)(void);
static pthread_mutex_t bounds_mutex_object;
static pthread_mutex_t bounds_mutex_malloc;
#endif

extern void (*__bounds_mutex_start)(void);
extern void (*__bounds_mutex_end)(void);

/* These routine are called to protect static data in the bounds_checking
   library.  */

static void
bounds_mutex_start(void)
{
#if USE_PTHREAD
    pthread_mutex_lock(&bounds_mutex_object);
#endif
}

static void
bounds_mutex_end(void)
{
#if USE_PTHREAD
    pthread_mutex_unlock(&bounds_mutex_object);
#endif
}

#if USE_PTHREAD
/* Protect malloc, realloc and free so multiple threads can use them.  */

static void *
bounds_malloc_hook(size_t size)
{
    void *r;

    if ((long)__realloc_hook == 0) {
        __malloc_hook = NULL;
        r = __bounds_malloc(size);
        __malloc_hook = bounds_malloc_hook;
    }
    else {
	pthread_mutex_lock(&bounds_mutex_malloc);
	__malloc_hook = NULL;
	r = __bounds_malloc(size);
	__malloc_hook = bounds_malloc_hook;
	pthread_mutex_unlock(&bounds_mutex_malloc);
    }
    return r;
}

static void *
bounds_realloc_hook(void *ptr, size_t size)
{
    void *r;

    pthread_mutex_lock(&bounds_mutex_malloc);
    __realloc_hook = NULL;
    r = __bounds_realloc(ptr, size);
    __realloc_hook = bounds_realloc_hook;
    pthread_mutex_unlock(&bounds_mutex_malloc);
    return r;
}

static void
bounds_free_hook(void *ptr)
{
    if ((long)__realloc_hook == 0) {
        __free_hook = NULL;
        __bounds_free(ptr);
        __free_hook = bounds_free_hook;
    }
    else {
	pthread_mutex_lock(&bounds_mutex_malloc);
	__free_hook = NULL;
	__bounds_free(ptr);
	__free_hook = bounds_free_hook;
	pthread_mutex_unlock(&bounds_mutex_malloc);
    }
}
#endif

/* Init routine to set up bounds_checking for the pthread library.  */

void
bounds_pthread_init(void)
{
#if USE_PTHREAD
    pthread_mutex_init(&bounds_mutex_object, NULL);
    pthread_mutex_init(&bounds_mutex_malloc, NULL);
    __bounds_function_thread = pthread_self;
    __realloc_hook = bounds_realloc_hook;
    __free_hook = bounds_free_hook;
    __malloc_hook = bounds_malloc_hook;
#endif
#if USE_SIGNAL
    __bounds_mutex_signal =  1;
#endif
    __bounds_mutex_start = bounds_mutex_start;
    __bounds_mutex_end = bounds_mutex_end;
}
#endif
